/*
Copyright (c) 2009, Christian Junker
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of the project nor the names of its contributors 
      may be used to endorse or promote products derived from this software
      without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY CHRISTIAN JUNKER ''AS IS'' AND ANY
EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL CHRISTIAN JUNKER BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 
*/

#include "DJAudioBuffer.h"

#include "DJLogger.h"

#include <string.h>

namespace DJ
{

template<typename T>
struct AudioFile::AudioBuffer<T>::D
{
    bool usefile;
    const AudioBuffer<T>* buf;
    const AudioFile* file;
    AudioBufferMode mode;
    unsigned short channels;

    CountT start;
    CountT size;
    CountT window;
    CountT windowpos;

    T*  data;

    Logger* log;
};

template<typename T>
AudioFile::AudioBuffer<T>::AudioBuffer(
    const AudioFile* file,
    const AudioBufferMode mode,
    const CountT start,
    const CountT size,
    const CountT window
)
{
    data = new D;
    data->log = Logger::GetInstancePtr();

    data->usefile = true;
    data->file = file;
    data->mode = mode;
    data->start = start;
    data->windowpos = start;
    data->size = size;
    data->channels = file->GetChannelCount();

    if(!window)
        data->window = size;
    else
        data->window = window;

    *data->log << "AudioBuffer: Initializing (size="<< data->size
              <<", window="<< data->window
              << ", channels="<< data->channels <<")"<<Log::NLn;

    data->data = new T[data->window];
    memset(data->data, 0, data->window);
    
}


template<typename T>
AudioFile::AudioBuffer<T>::AudioBuffer(
    const AudioBuffer<T>* buf,
    const CountT start,
    const CountT size,
    const CountT window
)
{
    data = new D;
    data->log = Logger::GetInstance();

    bool usefile =false;
    data->buf = buf;
    data->start = start;
    data->size = size;
    data->channels = buf->Channels();

    *data->log << "AudioBuffer: Initializing (size="<< size
              <<", window="<< window
              << ", channels="<< data->channels <<")"<<Log::NLn;

    if(!window)
        data->window = size;
    else
        data->window = window;

    data->data = 0;
}



template<typename T>
AudioFile::AudioBuffer<T>::~AudioBuffer()
{
    *data->log << "AudioBuffer: Destructing" << Log::NLn;
    delete data->data;
    delete data;
}

template<typename T>
const T&
AudioFile::AudioBuffer<T>::Get(
    const CountT at
) const
{
    CountT pos = at % data->size;

    if(data->usefile)
    {
        //load new window?
        bool load = at > data->windowpos + data->window ||
                    at < data->windowpos;

        //out of window, load new window
        if(load)
        {
            CountT start;
            start = at-(at%data->window);

            switch(data->mode)
            {
                case FRAMES:
                    data->file->ReadFrames(
                        start,
                        data->window/data->channels, // The array must be large enough to hold the product of frames and the number of channels.
                        data->data
                    );
                    break;

                case ITEMS:
                    data->file->Read(
                        start,
                        data->window,
                        data->data
                    );
                    break;
            }
            data->windowpos = start;
        }

        return data->data[at-data->windowpos];
    }
    else
    {
        return data->buf->Get(data->start + pos);
    }
}


template<typename T>
CountT
AudioFile::AudioBuffer<T>::Size() const
{
    return data->size;
}


template<typename T>
unsigned short
AudioFile::AudioBuffer<T>::Channels() const
{
    return data->channels;
}

template<typename T>
SampleRateT
AudioFile::AudioBuffer<T>::SampleRate() const
{
    if(data->usefile)
        return data->file->GetSampleRate();
    else
        return data->buf->SampleRate();
}

}
