/*
Copyright (c) 2009, Christian Junker
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of the project nor the names of its contributors 
      may be used to endorse or promote products derived from this software
      without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY CHRISTIAN JUNKER ''AS IS'' AND ANY
EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL CHRISTIAN JUNKER BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 
*/

#include "DJAudioFile.h"

#include "DJLogger.h"

#include <sndfile.h>
#include <assert.h>

#include <iostream>
namespace DJ
{


struct AudioFile::D
{
    std::string path;
    SNDFILE* file;
    SF_INFO  info;

    MetaData* meta;

    Logger* log;
};

AudioFile::AudioFile(
    std::string path
)
{
    data = new D;
    assert(path!="");
    data->path = path;

    data->log = Logger::GetInstancePtr();
    *data->log << "AudioFile:\tInitializing (path="<< path << ")" << Log::NLn;

    data->file = sf_open(data->path.c_str(), SFM_READ, &data->info);

    *data->log << "\tFramecount:\t" << data->info.frames << Log::NLn;
    *data->log << "\tSamplerate:\t" << data->info.samplerate << Log::NLn;
    *data->log << "\tChannels:\t" << data->info.channels << Log::NLn;
    *data->log << "\tSections:\t" << data->info.sections << Log::NLn;

    std::string seekable = data->info.seekable ? "yes" : "no";
    *data->log << "\tSeekable:\t" << seekable << Log::NLn;
}

AudioFile::~AudioFile()
{
    *data->log << "AudioFile: Destructing" << Log::NLn;

    int err = sf_close(data->file);
    if(err)
        *data->log <<"AudioFile:\tError - could not close file (Code=" << err <<")" << Log::NLn;

    data->log->finish();

    delete data;
}

///////////////
// Read Data //
///////////////
void
AudioFile::ReadFrames(const CountT start, const CountT frames, short* target) const
{
    sf_seek(data->file, start, SEEK_SET);
    sf_readf_short(data->file, target, frames);
}
void
AudioFile::ReadFrames(const CountT start, const CountT frames, int* target) const
{
    sf_seek(data->file, start, SEEK_SET);
    sf_readf_int(data->file, target, frames);
}
void
AudioFile::ReadFrames(const CountT start, const CountT frames, float* target) const
{
    sf_seek(data->file, start, SEEK_SET);
    sf_readf_float(data->file, target, frames);
}
void
AudioFile::ReadFrames(const CountT start, const CountT frames, double* target) const
{
    sf_seek(data->file, start, SEEK_SET);
    sf_readf_double(data->file, target, frames);
}


void
AudioFile::Read(const CountT start, const CountT frames, short* target) const
{
    sf_seek(data->file, start, SEEK_SET);
    sf_read_short(data->file, target, frames);
}
void
AudioFile::Read(const CountT start, const CountT frames, int* target) const
{
    sf_seek(data->file, start, SEEK_SET);
    sf_read_int(data->file, target, frames);
}
void
AudioFile::Read(const CountT start, const CountT frames, float* target) const
{
    sf_seek(data->file, start, SEEK_SET);
    sf_read_float(data->file, target, frames);
}
void
AudioFile::Read(const CountT start, const CountT frames, double* target) const
{
    sf_seek(data->file, start, SEEK_SET);
    sf_read_double(data->file, target, frames);
}


////////////
// GETTER //
////////////
std::string
AudioFile::GetPath() const
{
    return data->path;
}

const char*
AudioFile::GetFileTag(const AudioFileTag& tag) const
{
    return sf_get_string(data->file, tag);
}

CountT
AudioFile::GetFrameCount() const
{return data->info.frames;}

SampleRateT
AudioFile::GetSampleRate() const
{return data->info.samplerate;}

unsigned short
AudioFile::GetChannelCount() const
{return data->info.channels;}

FormatT
AudioFile::GetFormat() const
{return data->info.format;}

CountT
AudioFile::GetSectionCount() const
{return data->info.sections;}

bool
AudioFile::IsSeekable() const
{return data->info.seekable;}

}
