/*
Copyright (c) 2009, Christian Junker
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of the project nor the names of its contributors 
      may be used to endorse or promote products derived from this software
      without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY CHRISTIAN JUNKER ''AS IS'' AND ANY
EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL CHRISTIAN JUNKER BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 
*/

#include "DJExecTask.h"

#include "DJMP3Writer.h"
#include "DJQPlainTextLog.h"
#include "conf.h"

#include <QThread>
#include <QMutex>

#include <assert.h>

#include <QFileDialog>

namespace DJ
{

struct ExecTask::D
{
    std::string folder;
    ReadAudioFile* reader_a;
    ReadAudioFile* reader_b;
    ExtractMetaData* meta;
    unsigned short bitrate;

    QPushButton* lbl;

    MP3Writer* mp3a;
    MP3Writer* mp3b;
    MP3Writer* mp3;

    DoItThread* doit;

    bool cont;
    
    MP3Writer::Mode mode;
};

ExecTask::ExecTask() : QWidget()
{
    setupUi(this);

    data = new D;

    data->folder = "";
    data->reader_a = 0;
    data->reader_b = 0;
    data->meta = 0;
    data->bitrate = 320;
    data->mp3 = 0;
    data->mp3a = 0;
    data->mp3b = 0;
    data->cont = false;
    data->mode = MP3Writer::VBR;

    pteComment->setPlainText("Created with Disc-JoQey");

    data->doit = new DoItThread();

    connect(pbGo1, SIGNAL(clicked()), this, SLOT(StartA()));
    connect(pbGo2, SIGNAL(clicked()), this, SLOT(StartB()));
    connect(pbGoAll, SIGNAL(clicked()), this, SLOT(StartAll()));
    connect(this, SIGNAL(Start()), this, SLOT(Doit()), Qt::QueuedConnection);
    connect(data->doit, SIGNAL(Done()), this, SLOT(Cleanup()), Qt::QueuedConnection);
    connect(pbFolder, SIGNAL(clicked()), this, SLOT(Folder()));
    connect(this, SIGNAL(Done()), this, SLOT(Continue()), Qt::QueuedConnection);
    connect(rdCBR, SIGNAL(toggled(bool)), this, SLOT(SetMode(bool)));
}

ExecTask::~ExecTask()
{
    delete data;
}

void ExecTask::SetMode(bool vbr)
{
    if(vbr)
        data->mode = MP3Writer::CBR;
    else
        data->mode = MP3Writer::VBR;
}

void ExecTask::Folder()
{
    leFolder->setText(QFileDialog::getExistingDirectory(this, "Choose Output Directory", HOME_DIR));
}

void ExecTask::SetMeta(ExtractMetaData* meta)
{
    assert(meta);
    data->meta = meta;

    if(data->reader_a)
    {
        pbGo1->setEnabled(true);
        pbGoAll->setEnabled(true);
    }
    if(data->reader_b)
    {
        pbGo2->setEnabled(true);
        pbGoAll->setEnabled(true);
    }
}

void ExecTask::SetReaderA(ReadAudioFile* reader)
{
    assert(reader && data->meta);

    data->reader_a = reader;

    if(data->meta)
    {
        pbGo1->setEnabled(true);
        pbGoAll->setEnabled(true);
    }
}

void ExecTask::SetReaderB(ReadAudioFile* reader)
{
    assert(reader && data->meta);

    data->reader_b = reader;

    if(data->meta)
    {
        pbGo2->setEnabled(true);
        pbGoAll->setEnabled(true);
    }
}

void
ExecTask::StartAll()
{
    if(data->reader_a)
    {
        data->cont = true;
        StartA();
    }
    else if(data->reader_b)
    {
        data->cont = false;
        StartB();
    }
}

void
ExecTask::Continue()
{
    if(data->cont)
        StartB();
    data->cont = false;
}

void
ExecTask::StartA()
{
    leFolder->setEnabled(false);
    spnBitrate->setEnabled(false);
    pteComment->setEnabled(false);
    pbGo1->setEnabled(false);
    pbGo2->setEnabled(false);
    pbGoAll->setEnabled(false);

    pbGo1->setText("Working");
    data->lbl = pbGo1;

    if(data->mp3a)
        delete data->mp3a;

    data->mp3a = new MP3Writer(leFolder->text().toStdString(), data->meta, spnBitrate->value(), 1, data->mode);
    *data->reader_a >> data->mp3a;
    data->mp3 = data->mp3a;

    emit Start();
}

void
ExecTask::StartB()
{
    leFolder->setEnabled(false);
    spnBitrate->setEnabled(false);
    pteComment->setEnabled(false);
    pbGo1->setEnabled(false);
    pbGo2->setEnabled(false);
    pbGoAll->setEnabled(false);

    pbGo2->setText("Working");
    data->lbl = pbGo2;

    if(data->mp3b)
        delete data->mp3b;

    data->mp3b = new MP3Writer(leFolder->text().toStdString(), data->meta, spnBitrate->value(), 2, data->mode);
    *data->reader_b >> data->mp3b;
    data->mp3 = data->mp3b;

    emit Start();
}

void
ExecTask::Doit()
{
    if(!data->mp3)
        return;
    data->doit->Run(data->mp3);
}

void
ExecTask::Cleanup()
{
    leFolder->setEnabled(true);
    spnBitrate->setEnabled(true);
    pteComment->setEnabled(true);

    if(data->lbl == pbGo1)
        data->lbl->setText("Export Side A");
    else
        data->lbl->setText("Export Side B");

    if(data->reader_a)
    {
        pbGo1->setEnabled(true);
        pbGoAll->setEnabled(true);
    }
    if(data->reader_b)
    {
        pbGo2->setEnabled(true);
        pbGoAll->setEnabled(true);
    }

    emit Done();
}

}
