/*
Copyright (c) 2009, Christian Junker
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of the project nor the names of its contributors 
      may be used to endorse or promote products derived from this software
      without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY CHRISTIAN JUNKER ''AS IS'' AND ANY
EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL CHRISTIAN JUNKER BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 
*/

#include "DJMainWindow.h"

#include <QString>
#include <QRegExp>
#include <QMessageBox>
#include <QTableWidgetItem>

#include <sstream>

#include "conf.h"

#include "DJTrackListEditor.h"

#include "DJLogger.h"
#include "DJReadAudioFile.h"
#include "DJExtractAudioFrames.h"
#include "DJExtractMetaData.h"

#include "DJProjectTask.h"
#include "DJTracksTask.h"
#include "DJExecTask.h"

namespace DJ
{

class AudioFile;

struct MainWindow::D
{
    Logger* log;

    TracksTask* tracksTask;
    ExecTask* execTask;
};

MainWindow::MainWindow()
: QMainWindow()
{
    setupUi(this);

    pbPreprocessTask->setVisible(false);
    pbProjectTask->setVisible(false);

    data = new D;
    data->log = Logger::GetInstancePtr();
    *data->log << "MainWindow: Initializing" << Log::NLn;

    data->tracksTask = new TracksTask();
    data->execTask = new ExecTask();

    stckTaskArea->addWidget(data->tracksTask);
    stckTaskArea->addWidget(data->execTask);

    stckTaskArea->setCurrentIndex(0);

    SetupActions();

    data->tracksTask->Manage();
}

MainWindow::~MainWindow()
{
    *data->log << "MainWindow: Destructing" << Log::NLn;
    data->log->finish();

    delete data;
}

void
MainWindow::SetupActions()
{
    connect(pbTracksTask, SIGNAL(clicked()), this, SLOT(TracksTaskClicked()));
    connect(pbExecTask, SIGNAL(clicked()), this, SLOT(ExecTaskClicked()));

    connect(data->tracksTask, SIGNAL(PublishMeta(ExtractMetaData*)), data->execTask, SLOT(SetMeta(ExtractMetaData*)));
    connect(data->tracksTask, SIGNAL(PublishReaderA(ReadAudioFile*)), data->execTask, SLOT(SetReaderA(ReadAudioFile*)));
    connect(data->tracksTask, SIGNAL(PublishReaderB(ReadAudioFile*)), data->execTask, SLOT(SetReaderB(ReadAudioFile*)));

    connect(data->execTask, SIGNAL(Start()), this, SLOT(Activate()));
    connect(data->execTask, SIGNAL(Done()), this, SLOT(DeActivate()));

    connect(pbAbout, SIGNAL(clicked()), this, SLOT(About()));
}

void
MainWindow::About()
{
    QMessageBox::about(this, "About", "<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.0//EN\" \"http://www.w3.org/TR/REC-html40/strict.dtd\"><html><head><meta name=\"qrichtext\" content=\"1\" /><style type=\"text/css\">p, li { white-space: pre-wrap; }</style></head><body style=\" font-family:'Sans Serif'; font-size:10pt; font-weight:400; font-style:normal;\"><p align=\"center\" style=\" margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\"><img src=\":/logos/logojku.jpg\" /></p><p align=\"center\" style=\" margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\"><img src=\":/logos/cpbanner.jpg\" /></p><p style=\" margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\"><br /></p><p style=\" margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\">Disc-JoQey was developed by Christian Junker as a \"Practical (PR)\" at the Department of Computational Perception - Johannes Kepler University, Linz. Original idea and supervision by Peter Knees.</p><p style=\"-qt-paragraph-type:empty; margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\"></p><p style=\" margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\">Disc-JoQey is using the following Libraries:</p><p style=\" margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\">    * Qt</p><p style=\" margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\">    * libsndfile</p><p style=\" margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\">    * libmusicbrainz3</p><p style=\" margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\">    * libmp3lame</p><p style=\" margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\">    * id3lib</p><p style=\"-qt-paragraph-type:empty; margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\"></p><p style=\" margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\">Disc-JoQey is free software under the terms of the BSD-License. A copy is shipped with this distribution.</p></body></html>");
}
void
MainWindow::Activate()
{
    pbTracksTask->setEnabled(false);
    pbExecTask->setEnabled(false);
}

void
MainWindow::DeActivate()
{
    pbTracksTask->setEnabled(true);
    pbExecTask->setEnabled(true);
}


void
MainWindow::TracksTaskClicked()
{
    stckTaskArea->setCurrentWidget(data->tracksTask);
}

void
MainWindow::ExecTaskClicked()
{
    stckTaskArea->setCurrentWidget(data->execTask);
}

}
