/*
Copyright (c) 2009, Christian Junker
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of the project nor the names of its contributors 
      may be used to endorse or promote products derived from this software
      without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY CHRISTIAN JUNKER ''AS IS'' AND ANY
EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL CHRISTIAN JUNKER BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 
*/

#include "DJTracksTask.h"

#include "DJWaveFormInput.h"
#include "DJWaveFormMarker.h"
#include "DJAudioPlayer.h"
#include "DJExtractAudioFrames.h"

#include "DJTrackListEditor.h"

#include "DJOffsetGap.h"

#include <assert.h>
#include <sstream>

#include <iostream>
#include <math.h>
#include <stdlib.h>

#include <QMessageBox>

#include "conf.h"

#include "DJManageFiles.h"

namespace DJ
{
class ExtractMetaData;

struct TracksTask::D
{
    WaveFormInput<float>* winput_a;
    WaveFormScene<float>* scene_a;
    WaveFormMarker* marker_a;

    WaveFormInput<float>* winput_b;
    WaveFormScene<float>* scene_b;
    WaveFormMarker* marker_b;

    ReadAudioFile* reader_a;
    ReadAudioFile* reader_b;
    ExtractAudioFrames<float>* framebuffer_a;
    ExtractAudioFrames<float>* framebuffer_b;

    ExtractMetaData* extractmeta;

    AudioPlayer* player_a;
    AudioPlayer* player_b;

    TrackListEditor* track_edit;

    bool block_changed;
    bool action_setup;

    int side_focus;
};


TracksTask::TracksTask() : QWidget()
{
    setupUi(this);
    pbRefine->setVisible(false);
    pbManageFiles->setVisible(false);
    data = new D;

    data->block_changed = false;
    data->action_setup = true;

    data->reader_a = 0;
    data->reader_b = 0;

    data->side_focus = 1;

    data->extractmeta = new ExtractMetaData();

    data->track_edit = new TrackListEditor();
    tblTracks->setItemDelegate(data->track_edit);

    data->scene_a = new WaveFormScene<float>(1000, 145);
    gvSideA->setScene(data->scene_a);
    data->scene_b = new WaveFormScene<float>(1000, 145);
    gvSideB->setScene(data->scene_b);

    data->framebuffer_a = new ExtractAudioFrames<float>(100e3);
    data->winput_a = new WaveFormInput<float>(data->scene_a);
    data->marker_a = new WaveFormMarker(data->scene_a, 1); //side 1

    data->framebuffer_b = new ExtractAudioFrames<float>(100e3);
    data->winput_b = new WaveFormInput<float>(data->scene_b);
    data->marker_b = new WaveFormMarker(data->scene_b, 2); //side 2


    data->player_b = new AudioPlayer();
    data->player_a = new AudioPlayer();
}

TracksTask::~TracksTask()
{
    delete data;
}

void
TracksTask::SetupActions()
{

    //Track Task Actions
    connect(pbAddTrack, SIGNAL(clicked()), this, SLOT(AddTrack()));
    connect(pbRemTrack, SIGNAL(clicked()), this, SLOT(RemTrack()));
    connect(pbMvTrackUp, SIGNAL(clicked()), this, SLOT(MvTrackUp()));
    connect(pbMvTrackDown, SIGNAL(clicked()), this, SLOT(MvTrackDown()));
    connect(pbPlayTrack, SIGNAL(toggled(bool)), this, SLOT(PlayTrack(bool)));

    connect(data->player_a, SIGNAL(Tick(ms)), data->marker_a, SLOT(SetPos(ms)));
    connect(data->player_b, SIGNAL(Tick(ms)), data->marker_b, SLOT(SetPos(ms)));


    connect(tblTracks, SIGNAL(currentCellChanged(int,int,int,int)),
            this, SLOT(SelectTrack(int,int,int,int)));
    connect(tblTracks, SIGNAL(cellChanged(int,int)), this, SLOT(TrackEdit(int,int)));

    connect(data->extractmeta->GetOutput(), SIGNAL(MetaDataChanged(MetaDataTag,std::string)),
            this, SLOT(MetaDataChanged(MetaDataTag,std::string)));
    connect(data->extractmeta->GetOutput()->GetTrackList(), SIGNAL(TrackChanged(unsigned short)),
            this, SLOT(TrackListChanged(unsigned short)));
    connect(data->extractmeta->GetOutput()->GetTrackList(), SIGNAL(TrackRemoved(unsigned short)),
            this, SLOT(TrackRemoved(unsigned short)));
    connect(data->extractmeta->GetOutput(), SIGNAL(FocusChanged(short)),
            this, SLOT(FocusChanged(short)));

    connect(leArtist, SIGNAL(textEdited(QString)), data->extractmeta->GetOutput(), SLOT(SetArtist(QString)));
    connect(leAlbum, SIGNAL(textEdited(QString)), data->extractmeta->GetOutput(), SLOT(SetAlbum(QString)));
    connect(leGenre, SIGNAL(textEdited(QString)), data->extractmeta->GetOutput(), SLOT(SetGenre(QString)));
    connect(leYear, SIGNAL(textEdited(QString)), data->extractmeta->GetOutput(), SLOT(SetYear(QString)));

    connect(pbRefine, SIGNAL(clicked()), this, SLOT(Refine()));
    connect(pbRetrieve, SIGNAL(clicked()), this, SLOT(RetrieveMetaData()));

    connect(data->marker_a, SIGNAL(NewPos(ms)), data->player_a, SLOT(Seek(ms)));
    connect(data->marker_b, SIGNAL(NewPos(ms)), data->player_b, SLOT(Seek(ms)));

    connect(pbOffset, SIGNAL(clicked()), this, SLOT(ShowOffset()));
    connect(pbGap, SIGNAL(clicked()), this, SLOT(ShowGap()));

    connect(data->scene_a, SIGNAL(Selected()), this, SLOT(SideASelected()));
    connect(data->scene_b, SIGNAL(Selected()), this, SLOT(SideBSelected()));

}


void
TracksTask::SideASelected()
{
    data->side_focus = 1;
}

void
TracksTask::SideBSelected()
{
    data->side_focus = 2;
}

bool
StrEndsWith(std::string const &fullString, std::string const &ending)
{
    if (fullString.length() > ending.length()) {
        return (0 == fullString.compare (fullString.length() - ending.length(), ending.length(), ending));
    } else {
        return false;
    }
}

void
TracksTask::Manage()
{
    ManageFiles dlg;
    int ret = dlg.exec();
    if(ret != QDialog::Accepted)
        exit(0);

    std::string fa = dlg.GetFileA();
    std::string fb = dlg.GetFileB();

    if(fa == "" || fb == "")
        return;
    
    if(StrEndsWith(fa, ".mp3") || StrEndsWith(fb, ".mp3"))
        QMessageBox::information (0, "MP3 Import", "The MP3 File(s) are being imported. This may take a while. Press \"Ok\" to continue.");  
    
    if(StrEndsWith(fa, ".mp3"))
    {
        std::ostringstream tmp (std::ostringstream::out);
        tmp << TMP_DIR;
        tmp << rand() << ".wav";
        std::ostringstream cmd (std::ostringstream::out);
        cmd << DJ_LAME << " --decode \"" << fa << "\" \"" 
            << tmp.str() << "\"" << std::endl;
        system(cmd.str().data());
        fa = tmp.str();
    }
    
    if(StrEndsWith(fb, ".mp3"))
    {
        std::string tmp = "";
        do
        {
            std::ostringstream s (std::ostringstream::out);
            s << TMP_DIR << rand() << ".wav";
            tmp = s.str();
        } while(tmp == fa);
        std::ostringstream cmd (std::ostringstream::out);
        cmd << DJ_LAME << " --decode \"" << fb << "\" \"" 
            << tmp << "\"" << std::endl;
        system(cmd.str().data());
        fb = tmp;
    }

    if(data->reader_a)
        delete data->reader_a;
    if(data->reader_b)
        delete data->reader_b;

    data->reader_a = new ReadAudioFile(fa);
    data->reader_b = new ReadAudioFile(fb);

    assert(data->reader_a && data->reader_b && data->extractmeta);

    *data->reader_a >> data->extractmeta;
    data->extractmeta->Update();

    emit PublishMeta(data->extractmeta);
    emit PublishReaderA(data->reader_a);
    emit PublishReaderB(data->reader_b);

    //waveform -pipeline
    *data->reader_a >> data->framebuffer_a;
    *data->framebuffer_a >> data->winput_a;
    data->winput_a->Update();

    gvSideA->centerOn(0,0);

    *data->reader_b >> data->framebuffer_b;
    *data->framebuffer_b >> data->winput_b;
    data->winput_b->Update();

    gvSideB->centerOn(0,0);

    //mark tracks in waveform - pipeline
    *data->extractmeta >> data->marker_a;
    data->marker_a->Update();
    *data->extractmeta >> data->marker_b;
    data->marker_b->Update();

    //audioplayer pipeline;
    *data->reader_a >> data->player_a;
    data->player_a->Update();

    *data->reader_b >> data->player_b;
    data->player_b->Update();

    SetupActions();

}


void
TracksTask::ShowOffset()
{
    OffsetGap dlg;
    dlg.ShowDialog(OffsetGap::Offset, data->extractmeta->GetOutput()->AccessTrackList());
}

void
TracksTask::ShowGap()
{
    OffsetGap dlg;
    dlg.ShowDialog(OffsetGap::Gap, data->extractmeta->GetOutput()->AccessTrackList());
}


void
TracksTask::FocusChanged(short focus)
{
    tblTracks->setCurrentCell(focus, 0);
}

void
TracksTask::Refine()
{
    data->extractmeta->Refine();
}

void
TracksTask::RetrieveMetaData()
{
    this->setEnabled(false);
    this->setCursor(Qt::WaitCursor);
    SearchCriteria crit;
//     crit.Album = leAlbum->text().toUtf8().data();
//     crit.Artist = leArtist->text().toUtf8().data();
//     crit.Genre = leGenre->text().toUtf8().data();
    crit.Album = MetaData::QStringToStdString(leAlbum->text());
    crit.Artist = MetaData::QStringToStdString(leArtist->text());
    crit.Genre = MetaData::QStringToStdString(leGenre->text());
    crit.Length = 0;
    crit.Tracks = tblTracks->rowCount();
//     crit.Year = leYear->text().toUtf8().data();
    crit.Year = MetaData::QStringToStdString(leYear->text());

    data->extractmeta->Query((MetaDataEngine) cmbService->currentIndex(), &crit);

    this->setEnabled(true);
    this->unsetCursor();
}

void
TracksTask::AddTrack()
{
    TrackList* tracks = data->extractmeta->GetOutput()->AccessTrackList();

    TrackInfo newinf;
    if(!tracks->Size()) //empty
    {
        newinf.side = data->side_focus;
        newinf.start = 0;
        newinf.stop = 2000;
    }
    else
    {
        const TrackInfo* oldinf = tracks->End();
        newinf.side = data->side_focus;
        newinf.start = oldinf->stop;
        newinf.stop = oldinf->stop+2000;
    }
    std::ostringstream oss(std::ostringstream::out);
    oss << "Track " << tracks->Size()+1;
    newinf.title = oss.str();
    data->extractmeta->GetOutput()->AccessTrackList()->AddTrack(newinf);
}

void
TracksTask::MetaDataChanged(MetaDataTag tag, std::string val)
{
    int pos = 0;
    switch(tag)
    {
    case Album:
        pos = leAlbum->cursorPosition();
//         leAlbum->setText(QString::fromUtf8(val.data(), val.length()));
        leAlbum->setText(MetaData::StdStringToQString(val));
        leAlbum->setCursorPosition(pos);
        break;
    case Artist:
        pos = leArtist->cursorPosition();
//         leArtist->setText(QString::fromUtf8(val.data(), val.length()));
        leArtist->setText(MetaData::StdStringToQString(val));
        leArtist->setCursorPosition(pos);
        break;
    case Year:
        pos = leYear->cursorPosition();
//         leYear->setText(QString::fromUtf8(val.data(), val.length()));
        leYear->setText(MetaData::StdStringToQString(val));
        leYear->setCursorPosition(pos);
        break;
    case Genre:
        pos = leGenre->cursorPosition();
//         leGenre->setText(QString::fromUtf8(val.data(), val.length()));
        leGenre->setText(MetaData::StdStringToQString(val));
        leGenre->setCursorPosition(pos);
        break;
    default:
        break;
    }
}

void
TracksTask::TrackRemoved(unsigned short track)
{
    if(track<tblTracks->rowCount())
        tblTracks->removeRow(track);
}

void
TracksTask::TrackListChanged(unsigned short track)
{
    data->block_changed = true;
    if(track<tblTracks->rowCount())
    {
        tblTracks->blockSignals(true);
        const TrackInfo* inf = data->extractmeta->GetOutput()->GetTrackList()->At(track);
        
        tblTracks->item(track, 0)->setText(QString::number(inf->side));
        tblTracks->item(track, 1)->setText(QString::fromStdString(inf->title));
        
        ms start_ms_tot = inf->start%1000;
        ms start_ms = ((ms)round((double)start_ms_tot/100.))%10;
        ms start_ms_overflow = ((ms)round((double)start_ms_tot/100.))/10 * 1000;
        ms start_sek_tot = inf->start/1000 + start_ms_overflow;
        ms start_min = start_sek_tot/60;
        ms start_sek = start_sek_tot%60;
        QString start_txt = QString::number(start_min) + QString("m ")+QString::number(start_sek)+QString(",")+QString::number(start_ms)+QString("s");
        tblTracks->item(track, 2)->setText(start_txt);
        std::cout << start_txt.toStdString() << std::endl;
        
        ms stop_ms_tot = inf->stop%1000;
        ms stop_ms = ((ms)round((double)stop_ms_tot/100.))%10;
        ms stop_ms_overflow = ((ms)round((double)stop_ms_tot/100.))/10 * 1000;
        ms stop_sek_tot = inf->stop/1000 + stop_ms_overflow;
        ms stop_min = stop_sek_tot/60;
        ms stop_sek = stop_sek_tot%60;
        QString stop_txt = QString::number(stop_min) + QString("m ")+QString::number(stop_sek)+QString(",")+QString::number(stop_ms)+QString("s");
        tblTracks->item(track, 3)->setText(stop_txt);
        std::cout << stop_txt.toStdString() << std::endl;
        
        ms length_tot = inf->stop - inf->start;
        ms length_ms_tot = length_tot%1000;
        ms length_ms = ((ms)round((double)length_ms_tot/100.))%10;
        ms length_ms_overflow = ((ms)round((double)length_ms_tot/100.))/10 * 1000;
        ms length_sek = length_tot/1000 + length_ms_overflow;
        QString length_txt = QString::number(length_sek) + QString(",") + QString::number(length_ms) + QString("s");
        tblTracks->item(track, 4)->setText(length_txt);
        std::cout << length_txt.toStdString() << std::endl;

        if(inf->stop == inf->start)
        {
            tblTracks->item(track,2)->setBackgroundColor(Qt::red);
            tblTracks->item(track,3)->setBackgroundColor(Qt::red);
            tblTracks->item(track,4)->setBackgroundColor(Qt::red);
        }
        else
        {
            tblTracks->item(track,2)->setBackgroundColor(Qt::white);
            tblTracks->item(track,3)->setBackgroundColor(Qt::white);
            tblTracks->item(track,4)->setBackgroundColor(Qt::white);
        }
        tblTracks->blockSignals(false);
    }
    else if(track >= tblTracks->rowCount())
    {
        tblTracks->insertRow(tblTracks->rowCount());
        for(unsigned short i=0; i < tblTracks->columnCount(); i++)
            tblTracks->setItem(track, i, new QTableWidgetItem());
        TrackListChanged(track);
    }
    data->block_changed = false;
}

void
TracksTask::TrackEdit(int track, int field)
{
    if(data->block_changed)
        return;

    MetaData* meta = data->extractmeta->GetOutput();

    TrackInfo newinf;
    newinf.side = tblTracks->item(track,0)->text().toUShort();
    newinf.title = tblTracks->item(track,1)->text().toStdString();

    QStringList start_m_s = tblTracks->item(track,2)->text().split("m");
    if(start_m_s.size()<2) return;
    QStringList start_s_ms = start_m_s[1].split(",");
    if(start_s_ms.size()<2) return;
    QString start_m = start_m_s[0].replace(QRegExp("\\D"), "");
    QString start_s = start_s_ms[0].replace(QRegExp("\\D"), "");
    QString start_ms = start_s_ms[1].replace(QRegExp("\\D"), "");
    newinf.start = (start_m.toInt()*60*1000 + start_s.toInt()*1000 + start_ms.toInt()*100);
    
    if(field !=4)
    {
        QStringList stop_m_s = tblTracks->item(track,3)->text().split("m");
        if(stop_m_s.size()<2) return;
        QStringList stop_s_ms = stop_m_s[1].split(",");
        if(stop_s_ms.size()<2) return;
        QString stop_m = stop_m_s[0].replace(QRegExp("\\D"), "");
        QString stop_s = stop_s_ms[0].replace(QRegExp("\\D"), "");
        QString stop_ms = stop_s_ms[1].replace(QRegExp("\\D"), "");
        newinf.stop = (stop_m.toInt()*60*1000 + stop_s.toInt()*1000 + stop_ms.toInt()*100);
    }
    else
    {
        QStringList length_s_ms = tblTracks->item(track,4)->text().split(",");
        if(length_s_ms.size()<2) return;
        QString length_s = length_s_ms[0].replace(QRegExp("\\D"), "");
        QString length_ms = length_s_ms[1].replace(QRegExp("\\D"), "");
        ms length = length_s.toInt()*1000+length_ms.toInt()*100;
        newinf.stop = newinf.start + length;
    }

    if(newinf.stop <= newinf.start)
        newinf.stop = newinf.start;

    meta->AccessTrackList()->EditTrack(track, newinf);
    meta->SetFocus(track);
}

void
TracksTask::RemTrack()
{
    if(tblTracks->rowCount())
    {
        unsigned short track = tblTracks->currentRow();
        data->extractmeta->GetOutput()->AccessTrackList()->RemTrack(track);
    }
}

void
TracksTask::SelectTrack(int track,int,int,int)
{
    if(track >= 0)
        data->extractmeta->GetOutput()->SetFocus((short)track);
}

void
TracksTask::MvTrackUp()
{
    if(tblTracks->currentRow() > 0)
    {
        data->extractmeta->GetOutput()->AccessTrackList()->SwapTracks(tblTracks->currentRow(), tblTracks->currentRow()-1);
        data->extractmeta->GetOutput()->SetFocus(tblTracks->currentRow()-1);
    }
}

void
TracksTask::MvTrackDown()
{
    if(tblTracks->rowCount() >= 2 && tblTracks->currentRow() < tblTracks->rowCount()-1)
    {
        data->extractmeta->GetOutput()->AccessTrackList()->SwapTracks(tblTracks->currentRow(), tblTracks->currentRow()+1);
        data->extractmeta->GetOutput()->SetFocus(tblTracks->currentRow()+1);
    }
}

void
TracksTask::PlayTrack(bool play)
{
    MetaData* meta = data->extractmeta->GetOutput();
    short focus = meta->GetFocus();

    if(focus < 0)
    {
        if(pbPlayTrack->isChecked())
            pbPlayTrack->toggle();
        return;
    }

    if(!play)
    {
        pbAddTrack->setEnabled(true);
        pbRemTrack->setEnabled(true);
        pbMvTrackUp->setEnabled(true);
        pbMvTrackDown->setEnabled(true);
        pbRetrieve->setEnabled(true);
        pbRefine->setEnabled(true);
        tblTracks->setEnabled(true);
        pbPlayTrack->setText("Play");
        Stop();
    }
    else
    {
        pbAddTrack->setEnabled(false);
        pbRemTrack->setEnabled(false);
        pbMvTrackUp->setEnabled(false);
        pbMvTrackDown->setEnabled(false);
        pbRetrieve->setEnabled(false);
        pbRefine->setEnabled(false);
        tblTracks->setEnabled(false);
        pbPlayTrack->setText("Stop");
        Play(
            meta->GetTrackList()->At(focus)->start,
            meta->GetTrackList()->At(focus)->stop,
            meta->GetTrackList()->At(focus)->side
        );
    }
}

void
TracksTask::Play(ms start, ms stop, unsigned short side)
{
    switch(side)
    {
    case 1:
        data->player_a->Play(start, stop);
        break;
    case 2:
        data->player_b->Play(start, stop);
        break;
    default:
        return;
    }
}

void
TracksTask::Stop()
{
    data->player_a->Stop();
    data->player_b->Stop();
}

}
