/*
Copyright (c) 2009, Christian Junker
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of the project nor the names of its contributors 
      may be used to endorse or promote products derived from this software
      without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY CHRISTIAN JUNKER ''AS IS'' AND ANY
EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL CHRISTIAN JUNKER BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 
*/

#include "DJWaveFormMarker.h"

#include <QColor>
#include <QBrush>
#include <QPen>
#include <QFont>
#include <QCursor>

#include "DJTrackList.h"
#include "DJLogger.h"

#include <vector>
#include <assert.h>
#include <iostream>

#include <QGraphicsRectItem>

class QGraphicsLineItem;

namespace DJ
{

class WaveFormMarker::GraphicTrack :
    public QGraphicsRectItem
{
    public:
        GraphicTrack(
            WaveFormMarker* boss,
            unsigned short track,
            TrackList* tracks,
            MetaData* meta,
            WaveFormSceneCommon * scene,
            qreal x,
            qreal y,
            qreal w,
            qreal h,
            const QPen & pen = QPen(),
            const QBrush & brush = QBrush(),
            QGraphicsItem * parent = 0
        ) : QGraphicsRectItem(x,y,w,h,parent,scene)
        {
            this->setAcceptHoverEvents(true);
            this->setFlag(ItemIsSelectable,true);
            this->setPen(pen);
            this->setBrush(brush);
            dragging = false;
            this->tracks = tracks;
            this->track = track;
            this->meta = meta;
            this->boss = boss;
        }
        void SetTrack(unsigned short track)
        {
            this->track = track;
        }

        unsigned short GetTrack()
        {
            return track;
        }

    signals:
        void NewPos(ms);

    protected:
        virtual void mouseDoubleClickEvent ( QGraphicsSceneMouseEvent * event )
        {
            boss->PropagateNewPos(((WaveFormSceneCommon*) this->scene())->PosForPixel(event->pos().x()));
        }

        virtual QVariant itemChange ( GraphicsItemChange change, const QVariant & value )
        {
            if(change == QGraphicsItem::ItemSelectedChange && value.toBool())
            {
                if(meta->GetFocus() != track)
                    meta->SetFocus(track);
            }
            return value;
        }

        virtual void hoverMoveEvent ( QGraphicsSceneHoverEvent * event )
        {
            posinrect = event->pos().x() - this->rect().left();

            if( event->pos().x() > this->boundingRect().x() + this->boundingRect().width() -5)
            {
                left = false;
                dragging = true;
                this->setCursor(Qt::SizeHorCursor);
            }
            else if( event->pos().x() < this->boundingRect().x() + 5)
            {
                left = true;
                dragging = true;
                this->setCursor(Qt::SizeHorCursor);
            }
            else
            {
                dragging = false;
                this->setCursor(Qt::ArrowCursor);
            }
        }

        virtual void mouseMoveEvent ( QGraphicsSceneMouseEvent * event )
        {
            if(dragging)
            {
                ms n = ((WaveFormSceneCommon*) this->scene())->PosForPixel(event->pos().x());
                TrackInfo newinf;
                newinf.side = tracks->At(track)->side;
                newinf.start = tracks->At(track)->start;
                newinf.stop = tracks->At(track)->stop;

                newinf.title = tracks->At(track)->title;
                if(left)
                    if(n+1000 < newinf.stop)
                        newinf.start = n;
                    else
                        return;
                else
                    if(n > newinf.start +1000)
                        newinf.stop = n;
                    else
                        return;
                tracks->EditTrack(track, newinf);
            }
            else
            {
                this->setRect(event->pos().x()-posinrect,0,this->rect().width(),this->scene()->height());
                ms nstart = ((WaveFormSceneCommon*) this->scene())->PosForPixel(this->rect().left());
                ms nstop  = ((WaveFormSceneCommon*) this->scene())->PosForPixel(this->rect().right());
                TrackInfo newinf;
                newinf.side = tracks->At(track)->side;
                newinf.start = nstart;
                newinf.stop = nstop;
                newinf.title = tracks->At(track)->title;
                tracks->EditTrack(track, newinf);
            }
        }

    private:
        qreal posinrect;
        bool dragging;
        bool left;
        MetaData* meta;
        TrackList* tracks;
        unsigned short track;
        WaveFormMarker* boss;
};



struct WaveFormMarker::D
{
    Source< MetaData >* input;
    MetaData* meta;
    TrackList* tracks;
    unsigned short side;

    WaveFormSceneCommon* scene;
    short focus;

    qreal height;

    QGraphicsLineItem* cursor;

    QPen pe_track;
    QBrush br_track;
    std::vector< GraphicTrack* > graphical_tracks;

    Logger* log;
};

WaveFormMarker::WaveFormMarker(
    WaveFormSceneCommon* scene,
    const unsigned short side,
    QObject* parent
) : QObject(parent)
{
    assert(scene);

    data = new D;
    data->log = Logger::GetInstancePtr();
    *data->log << "WaveFormMarker:\tInitializing" << Log::NLn;

    data->input = 0;
    data->tracks = 0;
    data->meta = 0;
    data->side = side;
    data->focus = -1;

    data->scene = scene;
    QColor col_line(Qt::blue);
    QPen pe_line(col_line);
    pe_line.setWidth(2);
    data->cursor = scene->addLine(0,0,0,0,pe_line);
    data->cursor->setZValue(6);


    QColor col_track(Qt::blue);
    QPen pe(col_track);
    data->pe_track = pe;
    col_track.setAlpha(40);
    QBrush br(col_track);
    data->br_track = br;

    connect(data->scene, SIGNAL(sceneRectChanged(QRectF)), this, SLOT(Update(QRectF)));

    //add title
    QFont font;
    font.setBold(true);
    data->scene->addText("Side "+QString::number(side,10), font)->setZValue(10);
}

WaveFormMarker::~WaveFormMarker()
{
    data->scene->removeItem(data->cursor);
    delete data->cursor;

    delete data;
}

void
WaveFormMarker::SetInput(
    Source< MetaData >* input
)
{
    assert(input);
    data->input = input;
}

void
WaveFormMarker::Update(QRectF)
{
    unsigned short i;
    for(i=0; i<data->tracks->Size(); i++)
        UpdateTrack(i);
}

void
WaveFormMarker::Update()
{
    if(data->meta)
        disconnect(data->meta, 0, this, 0);
    if(data->tracks)
        disconnect(data->tracks, 0, this, 0);

    assert(data->input);
    data->meta = data->input->GetOutput();
    assert(data->meta);
    data->tracks = data->meta->AccessTrackList();
    assert(data->tracks);

    connect(data->meta, SIGNAL(FocusChanged(short)),
        this, SLOT(SetFocus(short)));
    connect(data->tracks, SIGNAL(TrackChanged(unsigned short)),
        this, SLOT(UpdateTrack(unsigned short)));
    connect(data->tracks, SIGNAL(TrackRemoved(unsigned short)),
        this, SLOT(RemoveTrack(unsigned short)));

    unsigned short i;
    for(i=0; i<data->tracks->Size(); i++)
        UpdateTrack(i);

    SetPos(0);
    SetFocus(0);
}

void
WaveFormMarker::SetPos(ms pos)
{
    static qreal p = 0;
    qreal np = data->scene->PixelForPos(pos);
    if(p != np)
    {
        p = np;
        data->cursor->setLine(p,0,p,data->scene->height());
    }
}

void
WaveFormMarker::PropagateNewPos(ms pos)
{
    emit NewPos(pos);
}

void
WaveFormMarker::UpdateTrack(unsigned short track)
{
    assert(data->tracks && track<data->tracks->Size());

    if(track >= data->graphical_tracks.size())
        data->graphical_tracks.resize(track+1, 0);

    if(data->tracks->At(track)->side != data->side)
    {
        if(data->graphical_tracks[track])
        {
            data->scene->removeItem(data->graphical_tracks[track]);
            delete data->graphical_tracks[track];
            data->graphical_tracks[track] = 0;
        }
        return;
    }

    if(!data->graphical_tracks[track])
    {
        data->graphical_tracks[track] = new GraphicTrack(this, track, data->tracks, data->meta, data->scene, 0,0,0,0,data->pe_track, data->br_track);
        data->graphical_tracks[track]->setZValue(6);
    }

    const TrackInfo* inf = data->tracks->At(track);
    qreal l = data->scene->PixelForPos( inf->start );
    qreal b = data->scene->height();
    qreal r = data->scene->PixelForPos( inf->stop );

    data->graphical_tracks[track]->setRect(l,0,r-l,b);
}

void
WaveFormMarker::RemoveTrack(unsigned short track)
{
    assert(data->tracks && track<data->graphical_tracks.size());

    for(unsigned i = 0; i<data->graphical_tracks.size(); i++)
        if(data->graphical_tracks[i])
            delete data->graphical_tracks[i];

    data->graphical_tracks.clear();

    for(unsigned i = 0; i<data->tracks->Size(); i++)
        UpdateTrack(i);
}

void
WaveFormMarker::SetFocus(short focus)
{
    assert(data->tracks);
    for(unsigned int i = 0; i<data->graphical_tracks.size(); i++)
        if(data->graphical_tracks[i])
            data->graphical_tracks[i]->setSelected(false);

    if(focus < 0 || focus >= (short) data->tracks->Size() ||
           data->tracks->At(focus)->side != data->side )
        return;

    assert(focus<=(short) data->tracks->Size());
    assert(data->graphical_tracks[focus]);
*data->log << "FOCUS " << focus << Log::NLn;
    data->graphical_tracks[focus]->setSelected(true);
}

}
