/*
Copyright (c) 2009, Christian Junker
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of the project nor the names of its contributors 
      may be used to endorse or promote products derived from this software
      without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY CHRISTIAN JUNKER ''AS IS'' AND ANY
EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL CHRISTIAN JUNKER BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 
*/

#include "DJWaveFormScene.h"

#include <QGraphicsTextItem>
#include <QGraphicsPathItem>
#include <QPainterPath>
#include <QPixmap>
#include <QPointF>

#include <QGraphicsSceneWheelEvent>

#include "DJLogger.h"

#include <assert.h>
#include <math.h>

#include <iostream>


namespace DJ
{

template <typename T>
struct WaveFormScene<T>::D
{
    unsigned short    channels;
    const AudioFile::AudioBuffer<T>*   amplitudes;
    CountT    count;
    CountT    perpixel;

    qreal   viewwidth;
    qreal   height;
    qreal   center;

    QPainterPath* path;
    QGraphicsPathItem* pathitem;

    Logger* log;
};


template <typename T>
WaveFormScene<T>::WaveFormScene(
    const qreal initWidth,
    const qreal initHeight,
    QObject* parent
) : WaveFormSceneCommon(parent)
{
    QGraphicsScene::QGraphicsScene();

    data = new D;
    data->log = Logger::GetInstancePtr();

    data->height = initHeight;
    data->viewwidth = initWidth;

    data->perpixel = 0;

    data->path = 0;
    data->pathitem = 0;

    data->amplitudes = 0;

    *data->log << "WaveFormScene:\tInitializing (initWidth=" << data->viewwidth
              << ", initHeight=" << data->height << ")" << Log::NLn;


    data->height -= 49;
    data->viewwidth -= 4;
    data->center = data->height/2.;
    *data->log << "\tExpanding Scene (width="<< data->viewwidth
              << ", height=" << data->height << ")" << Log::NLn;
    setSceneRect(0,0,data->viewwidth, data->height);
}


template <typename T>
WaveFormScene<T>::~WaveFormScene()
{
    *data->log << "WaveFormScene: Destructing" << Log::NLn;

    removeItem(data->pathitem);
    delete data->pathitem;
    delete data->path;

    data->log->finish();
    delete data;
}

//common
template<typename T>
qreal
WaveFormScene<T>::PixelForPos(const ms& pos) const
{
    if(!data->perpixel || !data->amplitudes)
        return 0;

    //pos in ms, samplerate in s
    return ((double)pos/1000.) * (double)data->amplitudes->SampleRate() / (double)data->perpixel / (double)data->channels;
}

template<typename T>
ms
WaveFormScene<T>::PosForPixel(const qreal& pixel) const
{
    if(!data->perpixel || !data->amplitudes)
        return 0;

    //pos in ms
    return ((double)data->perpixel * (double)pixel / (double)data->amplitudes->SampleRate() *data->channels)*1000.;
}


//events
template<typename T>
void
WaveFormScene<T>::wheelEvent (
    QGraphicsSceneWheelEvent * wheelEvent
)
{
    //standard 15° mouse -> delta is /120
    QRectF r = sceneRect();
    data->perpixel = ((double)data->count/(double)data->channels) / (double)(r.width() + wheelEvent->delta());
    setSceneRect(
        r.x(),
        r.y(),
        r.width() + wheelEvent->delta(),
        r.height()
    );

    ReDraw();
}

template <typename T>
void
WaveFormScene<T>::SetAmplitudes(
    const AudioFile::AudioBuffer<T>* amplitudes
)
{
    data->amplitudes = amplitudes;
    assert(data->amplitudes);
    data->count = data->amplitudes->Size();
    data->channels = data->amplitudes->Channels();

    *data->log << "WaveFormScene: Set amplitudes (count="<<data->count
              <<", channels="<<data->channels<<")"<<Log::NLn;
    data->perpixel = ((double)data->count/(double)data->channels) / (double)width();

    ReDraw();
}

template <typename T>
void
WaveFormScene<T>::ReDraw()
{
    *data->log << "WaveFormScene:\tRedrawing waveform" << Log::NLn;


    //drawing timeline
    addLine(0, data->center, width(), data->center)->setZValue(5);

    //drawing waveform
    ////remove old
    if(data->pathitem && data->path)
    {
        removeItem(data->pathitem);
        delete data->pathitem;
        delete data->path;
    }

    ////draw scene
    QPointF start(0,data->center);
    data->path = new QPainterPath(start);

    assert(data->count);
    if(data->amplitudes)
    {
        for(unsigned short c=0; c<1;++c)
        {
            CountT count = 0;
            T block_max = 0.;
            T block_min = 0.;
            for(CountT i = c; i<data->count; i+=data->channels)
            {
                int reread = 500;
                T val = data->amplitudes->Get(i);
                while(val > 1. || val < -1.) 
                {   
                    if(reread)
                        val = data->amplitudes->Get(i);
                    else
                    {
                        val = 0.;
                        break;
                    }
                    reread--;
                }
                block_max = val > block_max ? val : block_max;
                block_min = val < block_min ? val : block_min;
                val = data->amplitudes->Get(i+1);
                reread = 100;
                while(val > 1. || val < -1.) 
                {   
                    if(reread)
                        val = data->amplitudes->Get(i);
                    else
                    {
                        val = 0.;
                        break;
                    }
                    reread--;
                }
                block_max = val > block_max ? val : block_max;
                block_min = val < block_min ? val : block_min;
                if(count >= data->perpixel)
                {
                    data->path->lineTo(
                        data->path->currentPosition().x()+1,
                        data->center - (block_max*data->height*0.49)
                    );
                    data->path->lineTo(
                        data->path->currentPosition().x(),
                        data->center - (block_min*data->height*0.49)
                    );
                    block_max=0;
                    block_min=0;
                    count = 0;
                }
                else
                    ++count;
            }
        }
    }

    data->pathitem = new QGraphicsPathItem();
    data->pathitem->setPath(*data->path);
    data->pathitem->setZValue(3);
    data->pathitem->setPen(QPen(QColor(80, 190, 25)));
    addItem(data->pathitem);

    //done
    update();
}

}
