/*
Copyright (c) 2009, Christian Junker
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of the project nor the names of its contributors 
      may be used to endorse or promote products derived from this software
      without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY CHRISTIAN JUNKER ''AS IS'' AND ANY
EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL CHRISTIAN JUNKER BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 
*/

#include "DJMP3Writer.h"

#include "DJLog.h"

#include <lame/lame.h>
#include <id3/tag.h>
#include <id3/misc_support.h>

#include <cstring>

#include <assert.h>
#include <fstream>
#include <sstream>

namespace DJ
{

struct MP3Writer::D
{
    std::string folder;
    Source< AudioFile >* input;
    AudioFile*    file;
    ExtractMetaData* meta;
    unsigned short side;
    unsigned short bitrate;
    Logger* log;
    Mode mode;
};

MP3Writer::MP3Writer(std::string folder, ExtractMetaData* meta, unsigned short bitrate, unsigned short side, Mode mode)
{
    data = new D;
    data->log = Logger::GetInstancePtr();
    *data->log << "MP3Writer: Initializing" << Log::NLn;

    data->file = 0;
    data->input = 0;
    data->side = side;
    data->bitrate = bitrate;
    data->folder = folder;
    data->meta = meta;
    data->mode = mode;
}

MP3Writer::~MP3Writer()
{
    data->log->finish();
    delete data;
}

void
MP3Writer::SetInput(Source< AudioFile >* input)
{
    data->input = input;
}

void
MP3Writer::Update()
{
    assert(data->input);
    data->file = data->input->GetOutput();
    const TrackInfo* track = data->meta->GetOutput()->GetTrackList()->Begin();
    while(track && track <= data->meta->GetOutput()->GetTrackList()->End())
    {
        if(track->side == data->side)
            encode(track, data->folder + "/" + track->title + ".mp3");
        track++;
    }
}

void
MP3Writer::encode(
    const TrackInfo * track,
    const std::string& path
)
{
    assert(data->file);

    unsigned channels = data->file->GetChannelCount();
    int samplerate = data->file->GetSampleRate();
    unsigned long num_samples_channel = (double)(track->stop-track->start)/1000. * (double)samplerate;

    lame_global_flags* flags = lame_init();
    lame_set_num_samples(flags, num_samples_channel);
    lame_set_in_samplerate(flags, samplerate);
    lame_set_num_channels(flags, 2);
    lame_set_out_samplerate(flags, samplerate);
    lame_set_bWriteVbrTag(flags, 1);
    lame_set_quality(flags, 2);
    lame_set_mode(flags, STEREO);
    lame_set_original(flags, 1);
    lame_set_error_protection(flags, 1);
    if(data->mode == VBR)
    {
        lame_set_VBR(flags, vbr_abr);
        lame_set_VBR_q(flags, 0);
        lame_set_VBR_mean_bitrate_kbps(flags, data->bitrate);
        lame_set_VBR_min_bitrate_kbps(flags, .5*data->bitrate);
        lame_set_VBR_max_bitrate_kbps(flags, 2*data->bitrate);
    }
    else
    {
        lame_set_brate(flags, data->bitrate);
    }

    lame_set_write_id3tag_automatic(flags, 0);
    if(lame_init_params(flags) == -1)
        *data->log << "Error!" << Log::NLn;

    *data->log << "Encoding: " << track->title << Log::NLn;

    short *pcm = new short[num_samples_channel*channels];
    data->file->ReadFrames(
        track->start/1000. * samplerate,
        num_samples_channel,
        pcm
    );

    int mp3buf_size = 1.25*lame_get_num_samples(flags)+7200; //worst case
    unsigned char* mp3buf = new unsigned char[mp3buf_size];
    memset(mp3buf, 0, mp3buf_size);

    int written = lame_encode_buffer_interleaved(flags, pcm, num_samples_channel, mp3buf, mp3buf_size);
    lame_encode_flush(flags, mp3buf, written);

    lame_close(flags);
    std::ofstream out(path.c_str());
    unsigned char* p;
    for(p = mp3buf; p< mp3buf + written; p++)
        out << *p;

    out.close();

    //ID3
    MetaData* meta = data->meta->GetOutput();
    ID3_Tag myTag(path.c_str());

    ID3_AddArtist(&myTag, MetaData::StdStringToQString(meta->GetArtist()).toStdString().c_str(), true);
    std::ostringstream oss;
    oss << MetaData::StdStringToQString(meta->GetAlbum()).toStdString() << " Side " << track->side;
    ID3_AddAlbum(&myTag, oss.str().c_str(), true);
    ID3_AddGenre(&myTag, MetaData::StdStringToQString(meta->GetGenre()).toStdString().c_str(), true);
    ID3_AddTitle(&myTag, track->title.c_str(), true);
    ID3_AddTrack(&myTag, (unsigned char) 0, (unsigned char) meta->GetTrackList()->Size(), true);
    ID3_AddComment(&myTag, MetaData::StdStringToQString(meta->GetComment()).toStdString().c_str(), true);
    ID3_AddYear(&myTag, MetaData::StdStringToQString(meta->GetYear()).toStdString().c_str(), true);

    myTag.Update();


    delete pcm;
    delete mp3buf;
}

}
