/*
Copyright (c) 2009, Christian Junker
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of the project nor the names of its contributors 
      may be used to endorse or promote products derived from this software
      without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY CHRISTIAN JUNKER ''AS IS'' AND ANY
EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL CHRISTIAN JUNKER BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 
*/

#include "DJLogger.h"

#include <vector>

#include <QMutex>
#include <QWaitCondition>

#include <sstream>

using std::vector;
using std::string;

namespace DJ
{

struct Logger::D
{
    std::vector<std::string>    *MsgQueue;

    QReadWriteLock  MsgMutex;
    QMutex          Wait;
    QWaitCondition  Notify;
};

unsigned Logger::m_Ref = 0;

const std::string Logger::m_NewLine = "___NEW_LINE___";

vector<Log *> *Logger::m_Logs = new vector<Log *>();

QReadWriteLock  Logger::m_LogMutex;
Logger          *Logger::m_Self;

Logger::Logger()
{
    data = new D;
    data->MsgQueue = new vector<string>();
}

Logger::~Logger()
{
    data->Notify.wakeAll();
    m_Self->wait();

    delete data->MsgQueue;
    m_Self = 0;

    delete data;
}

void
Logger::finish()
{
    --m_Ref;

    if(!m_Ref)
    {
        delete m_Self;
    }
}


//static
Logger&
Logger::GetInstance()
{
    ++m_Ref;

    if(m_Self == NULL)
    {
        m_Self = new Logger();
        m_Self->start();
    }

    return *m_Self;
}

//static
Logger*
Logger::GetInstancePtr()
{
    ++m_Ref;

    if(m_Self == NULL)
    {
        m_Self = new Logger();
        m_Self->start();
    }

    return m_Self;
}


//static
unsigned
Logger::AddOutput(
    Log * out
)
{
    m_LogMutex.lockForWrite();

    unsigned id = m_Logs->size();

    m_Logs->push_back(out);

    m_LogMutex.unlock();

    return id;
}


//static
void
Logger::RemOutput(
    const unsigned& id,
    bool del
)
{
    m_LogMutex.lockForWrite();

    if(m_Logs->size() > id)
    {
        if(del)
            delete m_Logs->at(id); //also del ref

        m_Logs->erase(m_Logs->begin()+id);
    }

    m_LogMutex.unlock();
}


//static
void
Logger::RemOutput(
    Log * out,
    bool del
)
{
    vector<Log *>::iterator it = m_Logs->begin();
    while(it != m_Logs->end())
    {
        if(*it == out)
            RemOutput(it - m_Logs->begin(), del);

        it++;
    }
}


template<typename T>
std::string
Logger::Prepare(
    const T& val
) const
{
    std::ostringstream oss;
    oss << val;
    return oss.str();
}


Logger& 
Logger::operator<<(
    string msg
)
{
     if(msg == "") //nothing todo
        return *this;

    data->MsgMutex.lockForWrite();
    if(msg.find_first_of('\n') == std::string::npos) //proper string
    {
        data->MsgQueue->push_back(msg);
        data->MsgMutex.unlock();
        data->Notify.wakeAll();
        return *this;
    }

    int pos = 0;
    while( (pos=msg.find_first_of('\n')) != (int) std::string::npos )
    {
        data->MsgQueue->push_back( msg.substr(0, pos) );
        data->MsgQueue->push_back( m_NewLine );
        if(pos != (int) msg.size()-1)
            msg = msg.substr(pos+1);
        else
            msg = "";
    }
    if(msg != "")
        data->MsgQueue->push_back( msg );

    data->MsgMutex.unlock();
    data->Notify.wakeAll();
    return *this;
}

void
Logger::run()
{
    while(m_Ref)
    {
        //wait till a message arrived
        if(data->MsgQueue->empty())
        {
            data->Wait.lock();
            data->Notify.wait(&data->Wait);
            data->Wait.unlock();
        }

        //print messages to streams
        while(!data->MsgQueue->empty())
        {
            data->MsgMutex.lockForRead();
            const string msg = data->MsgQueue->front();
            data->MsgQueue->erase(data->MsgQueue->begin());
            data->MsgMutex.unlock();

            m_LogMutex.lockForRead();
            vector<Log *>::iterator s_it = m_Logs->begin();
            while(s_it != m_Logs->end())
            {
                if(msg != m_NewLine)
                    *(*s_it) <<  msg;
                else
                    *(*s_it) <<  Log::NLn;
                s_it++;
            }
            m_LogMutex.unlock();
        }
    }
}


Logger& Logger::operator<< (bool& val)
{ *this<< Prepare<bool>( val ); return *this; }

Logger& Logger::operator<< (short& val)
{ *this<< Prepare<short>( val ); return *this; }

Logger& Logger::operator<< (unsigned short& val)
{*this<< Prepare<unsigned short>( val ); return *this; }

Logger& Logger::operator<< (int& val)
{*this<< Prepare<int>( val ); return *this; }

Logger& Logger::operator<< (unsigned int& val)
{*this<< Prepare<unsigned int>( val ); return *this; }

Logger& Logger::operator<< (long& val)
{*this<< Prepare<long>( val ); return *this; }

Logger& Logger::operator<< (unsigned long& val)
{*this<< Prepare<unsigned long>( val ); return *this; }

Logger& Logger::operator<< (float& val)
{*this<< Prepare<float>( val ); return *this; }

Logger& Logger::operator<< (double& val)
{*this<< Prepare<double>( val ); return *this; }

Logger& Logger::operator<< (long double& val)
{*this<< Prepare<long double>( val ); return *this; }

Logger& Logger::operator<< (long long& val)
{*this<< Prepare<long long>( val ); return *this; }
}
