/*
Copyright (c) 2009, Christian Junker
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of the project nor the names of its contributors 
      may be used to endorse or promote products derived from this software
      without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY CHRISTIAN JUNKER ''AS IS'' AND ANY
EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL CHRISTIAN JUNKER BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 
*/

#ifndef DJ_PIPELINE_H
#define DJ_PIPELINE_H

#include "DJArray.h"
#include "DJNoMultithreadingException.h"

namespace DJ
{

template<typename T>
class Sink;

/**
 *  \class Source
 *  \brief A pipeline source
 *
 *  This is the base class for true sources
 *  in the pipeline. Sources can only produce.
 *  |->
 */
template< typename T >
class Source
{
    public:
        Source() {}
        virtual ~Source() {}

        /**
         * \brief Connect the source to a sink
         */
        void operator >> (Sink<T>* out);

        /**
         * \brief Retrieve output
         */
        virtual T* GetOutput() = 0;

        /**
         * \brief Retrieve multithreaded output
         */
        virtual Array<T>* GetOutputs()
            throw(NoMultiThreadingException);

    protected:
        /**
         * \brief The implementation of the connect operator
         */
        virtual void connect(Sink<T>* out);

    private:
        Source(const Source&); //purposely not implemented
        void operator=(const Source&); //purposely not implemented
};


/**
 *  \class Sink
 *  \brief A pipeline sink
 *
 *  This is the base class for true sinks
 *  in the pipeline. Sinks can only receive.
 *  ->|
 */
template< typename T >
class Sink
{
    public:
        Sink() {}
        virtual ~Sink() {}

        /**
         * \brief Set input
         */
        virtual void SetInput(Source<T>*) = 0;

        /**
         * \brief Set multithreaded input
         */
        virtual void SetInput(Source< Array<T> >* )
            throw(NoMultiThreadingException);

        /**
         * \brief Update data
         * if not implemented the remaining path is not informed about updates.
         * this is a good idea for pure passthroughs.
         */
        virtual void Update() {}

    private:
        Sink(const Sink&); //purposely not implemented
        void operator=(const Sink&); //purposely not implemented
};


/**
 *  \class Filter
 *  \brief A pipeline element
 *
 *  This is the base class for pipeline elements
 *  which act both as a sink and a source.
 *  they receive an input, perform an action to
 *  produce an output.
 *  ->|->
 */
template< typename T_in, typename T_out=T_in>
class Filter :
    public Sink<T_in>,
    public Source<T_out>
{
    public:
        Filter() {}
        virtual ~Filter() {}

        /**
         * \brief Retrieve output
         */
        virtual T_out* GetOutput() = 0;

        /**
         * \brief Set input
         */
        virtual void SetInput(Source<T_in>*) = 0;

        /**
         * \brief Update data
         */
        virtual void Update() {}

    private:
        Filter(const Filter&); //purposely not implemented
        void operator=(const Filter&); //purposely not implemented
};


/**
 *  \class Splitter
 *  \brief Multithreading data
 *
 *  The splitter is a special case of a filter.
 *  It is used to split data in order to use
 *  it in filters supporting multithreading.
 *    |->
 *  ->|->
 *    |->
 */
template< typename T>
class Splitter :
    public Filter< T, Array<T> >
{
    public:
        Splitter() {}
        virtual ~Splitter() {}

    private:
        Splitter(const Splitter&); //purposely not implemented
        void operator=(const Splitter&); //purposely not implemented
};

/**
 *  \class Joiner
 *  \brief Sequentializing data
 *
 *  The Joiner is a special case of a filter.
 *  It is used to combine data in order to use
 *  it in single threaded filters.
 *  ->|
 *  ->|->
 *  ->|
 */
template< typename T>
class Joiner :
    public Filter< Array<T>, T >
{
    public:
        Joiner() {}
        virtual ~Joiner() {}

    private:
        Joiner(const Joiner&); //purposely not implemented
        void operator=(const Joiner&); //purposely not implemented
};

}

#ifndef DJ_MANUAL_INSTANTIATION
    #include "DJPipeline.txx"
#endif

#endif
